@@ -24,15 +24,15 @@
 #
 ###############################################################################
 
-from __future__ import absolute_import
-
 import os
-import six
+import re
 import struct
 import platform
 import math
+import decimal
+from binascii import b2a_hex, a2b_hex
 
-from autobahn.util import time_ns
+from txaio import time_ns
 from autobahn.wamp.interfaces import IObjectSerializer, ISerializer
 from autobahn.wamp.exception import ProtocolError
 from autobahn.wamp import message
@@ -279,7 +279,7 @@ class Serializer(object):
         except Exception as e:
             raise ProtocolError("invalid serialization of WAMP message: {0} {1}".format(type(e).__name__, e))
 
-        if self._serializer.NAME == u'flatbuffers':
+        if self._serializer.NAME == 'flatbuffers':
             msgs = raw_msgs
         else:
             msgs = []
@@ -289,13 +289,11 @@ class Serializer(object):
                     raise ProtocolError("invalid type {0} for WAMP message".format(type(raw_msg)))
 
                 if len(raw_msg) == 0:
-                    raise ProtocolError(u"missing message type in WAMP message")
+                    raise ProtocolError("missing message type in WAMP message")
 
                 message_type = raw_msg[0]
 
-                if type(message_type) not in six.integer_types:
-                    # CBOR doesn't roundtrip number types
-                    # https://bitbucket.org/bodhisnarkva/cbor/issues/6/number-types-dont-roundtrip
+                if type(message_type) != int:
                     raise ProtocolError("invalid type {0} for WAMP message type".format(type(message_type)))
 
                 Klass = self.MESSAGE_TYPE_MAP.get(message_type)
@@ -314,7 +312,7 @@ class Serializer(object):
         self._unserialized_rated_messages += int(math.ceil(float(len(payload)) / self.RATED_MESSAGE_SIZE))
 
         # maybe auto-reset and trigger user callback ..
-        if self._autoreset_callback and ((self._autoreset_duration and (time_ns() - self._stats_reset) >= self._autoreset_duration) or(self._autoreset_rated_messages and self.stats_rated_messages() >= self._autoreset_rated_messages)):
+        if self._autoreset_callback and ((self._autoreset_duration and (time_ns() - self._stats_reset) >= self._autoreset_duration) or (self._autoreset_rated_messages and self.stats_rated_messages() >= self._autoreset_rated_messages)):
             stats = self.stats(reset=True)
             self._autoreset_callback(stats)
 
@@ -346,9 +344,22 @@ else:
 
     class _WAMPJsonEncoder(json.JSONEncoder):
 
+        def __init__(self, *args, **kwargs):
+            if 'use_binary_hex_encoding' in kwargs:
+                self._use_binary_hex_encoding = kwargs['use_binary_hex_encoding']
+                del kwargs['use_binary_hex_encoding']
+            else:
+                self._use_binary_hex_encoding = False
+            json.JSONEncoder.__init__(self, *args, **kwargs)
+
         def default(self, obj):
-            if isinstance(obj, six.binary_type):
-                return u'\x00' + base64.b64encode(obj).decode('ascii')
+            if isinstance(obj, bytes):
+                if self._use_binary_hex_encoding:
+                    return '0x' + b2a_hex(obj).decode('ascii')
+                else:
+                    return '\x00' + base64.b64encode(obj).decode('ascii')
+            elif isinstance(obj, decimal.Decimal):
+                return str(obj)
             else:
                 return json.JSONEncoder.default(self, obj)
 
@@ -359,16 +370,51 @@ else:
     from json import scanner
     from json.decoder import scanstring
 
-    def _parse_string(*args, **kwargs):
-        s, idx = scanstring(*args, **kwargs)
-        if s and s[0] == u'\x00':
-            s = base64.b64decode(s[1:])
-        return s, idx
+    _DEC_MATCH = re.compile(r'^[\+\-E\.0-9]+$')
 
     class _WAMPJsonDecoder(json.JSONDecoder):
 
         def __init__(self, *args, **kwargs):
+            if 'use_binary_hex_encoding' in kwargs:
+                self._use_binary_hex_encoding = kwargs['use_binary_hex_encoding']
+                del kwargs['use_binary_hex_encoding']
+            else:
+                self._use_binary_hex_encoding = False
+
+            if 'use_decimal_from_str' in kwargs:
+                self._use_decimal_from_str = kwargs['use_decimal_from_str']
+                del kwargs['use_decimal_from_str']
+            else:
+                self._use_decimal_from_str = False
+
+            if 'use_decimal_from_float' in kwargs:
+                self._use_decimal_from_float = kwargs['use_decimal_from_float']
+                del kwargs['use_decimal_from_float']
+                if self._use_decimal_from_float:
+                    kwargs['parse_float'] = decimal.Decimal
+            else:
+                self._use_decimal_from_str = False
+
             json.JSONDecoder.__init__(self, *args, **kwargs)
+
+            def _parse_string(*args, **kwargs):
+                s, idx = scanstring(*args, **kwargs)
+                if self._use_binary_hex_encoding:
+                    if s and s[0:2] == '0x':
+                        s = a2b_hex(s[2:])
+                        return s, idx
+                else:
+                    if s and s[0] == '\x00':
+                        s = base64.b64decode(s[1:])
+                        return s, idx
+                if self._use_decimal_from_str and _DEC_MATCH.match(s):
+                    try:
+                        s = decimal.Decimal(s)
+                        return s, idx
+                    except decimal.InvalidOperation:
+                        pass
+                return s, idx
+
             self.parse_string = _parse_string
 
             # we need to recreate the internal scan function ..
@@ -378,14 +424,19 @@ else:
             # not the C version, as the latter won't work
             # self.scan_once = scanner.make_scanner(self)
 
-    def _loads(s):
-        return json.loads(s, cls=_WAMPJsonDecoder)
+    def _loads(s, use_binary_hex_encoding=False, use_decimal_from_str=False, use_decimal_from_float=False):
+        return json.loads(s,
+                          use_binary_hex_encoding=use_binary_hex_encoding,
+                          use_decimal_from_str=use_decimal_from_str,
+                          use_decimal_from_float=use_decimal_from_float,
+                          cls=_WAMPJsonDecoder)
 
-    def _dumps(obj):
+    def _dumps(obj, use_binary_hex_encoding=False):
         return json.dumps(obj,
                           separators=(',', ':'),
                           ensure_ascii=False,
                           sort_keys=False,
+                          use_binary_hex_encoding=use_binary_hex_encoding,
                           cls=_WAMPJsonEncoder)
 
     _json = json
@@ -398,25 +449,35 @@ class JsonObjectSerializer(object):
     The JSON module used (now only stdlib).
     """
 
-    NAME = u'json'
+    NAME = 'json'
 
     BINARY = False
 
-    def __init__(self, batched=False):
+    def __init__(self, batched=False, use_binary_hex_encoding=False, use_decimal_from_str=False, use_decimal_from_float=False):
         """
-        Ctor.
 
         :param batched: Flag that controls whether serializer operates in batched mode.
         :type batched: bool
+
+        :param use_binary_hex_encoding: Flag to enable HEX encoding prefixed with ``"0x"``,
+            otherwise prefix binaries with a ``\0`` byte.
+        :type use_binary_hex_encoding: bool
+
+        :param use_decimal_from_str: Flag to automatically encode Decimals as strings, and
+            to try to parse strings as Decimals.
+        :type use_decimal_from_str: bool
         """
         self._batched = batched
+        self._use_binary_hex_encoding = use_binary_hex_encoding
+        self._use_decimal_from_str = use_decimal_from_str
+        self._use_decimal_from_float = use_decimal_from_float
 
     def serialize(self, obj):
         """
         Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.serialize`
         """
-        s = _dumps(obj)
-        if isinstance(s, six.text_type):
+        s = _dumps(obj, use_binary_hex_encoding=self._use_binary_hex_encoding)
+        if isinstance(s, str):
             s = s.encode('utf8')
         if self._batched:
             return s + b'\30'
@@ -433,7 +494,10 @@ class JsonObjectSerializer(object):
             chunks = [payload]
         if len(chunks) == 0:
             raise Exception("batch format error")
-        return [_loads(data.decode('utf8')) for data in chunks]
+        return [_loads(data.decode('utf8'),
+                       use_binary_hex_encoding=self._use_binary_hex_encoding,
+                       use_decimal_from_str=self._use_decimal_from_str,
+                       use_decimal_from_float=self._use_decimal_from_float) for data in chunks]
 
 
 IObjectSerializer.register(JsonObjectSerializer)
@@ -442,7 +506,7 @@ SERID_TO_OBJSER[JsonObjectSerializer.NAME] = JsonObjectSerializer
 
 class JsonSerializer(Serializer):
 
-    SERIALIZER_ID = u"json"
+    SERIALIZER_ID = "json"
     """
     ID used as part of the WebSocket subprotocol name to identify the
     serializer with WAMP-over-WebSocket.
@@ -454,22 +518,24 @@ class JsonSerializer(Serializer):
     handshake identify the serializer with WAMP-over-RawSocket.
     """
 
-    MIME_TYPE = u"application/json"
+    MIME_TYPE = "application/json"
     """
     MIME type announced in HTTP request/response headers when running
     WAMP-over-Longpoll HTTP fallback.
     """
 
-    def __init__(self, batched=False):
+    def __init__(self, batched=False, use_binary_hex_encoding=False, use_decimal_from_str=False):
         """
         Ctor.
 
         :param batched: Flag to control whether to put this serialized into batched mode.
         :type batched: bool
         """
-        Serializer.__init__(self, JsonObjectSerializer(batched=batched))
+        Serializer.__init__(self, JsonObjectSerializer(batched=batched,
+                                                       use_binary_hex_encoding=use_binary_hex_encoding,
+                                                       use_decimal_from_str=use_decimal_from_str))
         if batched:
-            self.SERIALIZER_ID = u"json.batched"
+            self.SERIALIZER_ID = "json.batched"
 
 
 ISerializer.register(JsonSerializer)
@@ -477,7 +543,7 @@ SERID_TO_SER[JsonSerializer.SERIALIZER_ID] = JsonSerializer
 
 
 _HAS_MSGPACK = False
-_USE_UMSGPACK = platform.python_implementation() == u'PyPy' or 'AUTOBAHN_USE_UMSGPACK' in os.environ
+_USE_UMSGPACK = platform.python_implementation() == 'PyPy' or 'AUTOBAHN_USE_UMSGPACK' in os.environ
 
 if not _USE_UMSGPACK:
     try:
@@ -513,7 +579,7 @@ if _HAS_MSGPACK:
 
     class MsgPackObjectSerializer(object):
 
-        NAME = u'msgpack'
+        NAME = 'msgpack'
 
         MSGPACK_MODULE = _msgpack
 
@@ -581,7 +647,7 @@ if _HAS_MSGPACK:
 
     class MsgPackSerializer(Serializer):
 
-        SERIALIZER_ID = u"msgpack"
+        SERIALIZER_ID = "msgpack"
         """
         ID used as part of the WebSocket subprotocol name to identify the
         serializer with WAMP-over-WebSocket.
@@ -593,7 +659,7 @@ if _HAS_MSGPACK:
         handshake identify the serializer with WAMP-over-RawSocket.
         """
 
-        MIME_TYPE = u"application/x-msgpack"
+        MIME_TYPE = "application/x-msgpack"
         """
         MIME type announced in HTTP request/response headers when running
         WAMP-over-Longpoll HTTP fallback.
@@ -608,7 +674,7 @@ if _HAS_MSGPACK:
             """
             Serializer.__init__(self, MsgPackObjectSerializer(batched=batched))
             if batched:
-                self.SERIALIZER_ID = u"msgpack.batched"
+                self.SERIALIZER_ID = "msgpack.batched"
 
     ISerializer.register(MsgPackSerializer)
     SERID_TO_SER[MsgPackSerializer.SERIALIZER_ID] = MsgPackSerializer
@@ -617,39 +683,31 @@ if _HAS_MSGPACK:
 
 
 _HAS_CBOR = False
-if 'AUTOBAHN_USE_CBOR2' in os.environ:
-    try:
-        # https://pypi.org/project/cbor2/
-        # https://github.com/agronholm/cbor2
-        import cbor2
-    except ImportError:
-        pass
-    else:
-        _HAS_CBOR = True
-        _cbor_loads = cbor2.loads
-        _cbor_dumps = cbor2.dumps
-        _cbor = cbor2
-        # print('Notice: Autobahn is using cbor2 library for CBOR serialization')
+
+
+try:
+    import cbor2
+except ImportError:
+    pass
 else:
-    try:
-        # https://pypi.python.org/pypi/cbor
-        # https://bitbucket.org/bodhisnarkva/cbor
-        import cbor
-    except ImportError:
-        pass
-    else:
-        _HAS_CBOR = True
-        _cbor_loads = cbor.loads
-        _cbor_dumps = cbor.dumps
-        _cbor = cbor
-        # print('Notice: Autobahn is using cbor library for CBOR serialization')
+    _HAS_CBOR = True
+    _cbor_loads = cbor2.loads
+    _cbor_dumps = cbor2.dumps
+    _cbor = cbor2
 
 
 if _HAS_CBOR:
 
     class CBORObjectSerializer(object):
+        """
+        CBOR serializer based on `cbor2 <https://github.com/agronholm/cbor2>`_.
 
-        NAME = u'cbor'
+        This CBOR serializer has proper support for arbitrary precision decimals,
+        via tagged decimal fraction encoding, as described in
+        `RFC7049 section 2.4.3 <https://datatracker.ietf.org/doc/html/rfc7049#section-2.4.3>`_.
+        """
+
+        NAME = 'cbor'
 
         CBOR_MODULE = _cbor
 
@@ -660,7 +718,6 @@ if _HAS_CBOR:
 
         def __init__(self, batched=False):
             """
-            Ctor.
 
             :param batched: Flag that controls whether serializer operates in batched mode.
             :type batched: bool
@@ -718,7 +775,7 @@ if _HAS_CBOR:
 
     class CBORSerializer(Serializer):
 
-        SERIALIZER_ID = u"cbor"
+        SERIALIZER_ID = "cbor"
         """
         ID used as part of the WebSocket subprotocol name to identify the
         serializer with WAMP-over-WebSocket.
@@ -730,7 +787,7 @@ if _HAS_CBOR:
         handshake identify the serializer with WAMP-over-RawSocket.
         """
 
-        MIME_TYPE = u"application/cbor"
+        MIME_TYPE = "application/cbor"
         """
         MIME type announced in HTTP request/response headers when running
         WAMP-over-Longpoll HTTP fallback.
@@ -745,7 +802,7 @@ if _HAS_CBOR:
             """
             Serializer.__init__(self, CBORObjectSerializer(batched=batched))
             if batched:
-                self.SERIALIZER_ID = u"cbor.batched"
+                self.SERIALIZER_ID = "cbor.batched"
 
     ISerializer.register(CBORSerializer)
     SERID_TO_SER[CBORSerializer.SERIALIZER_ID] = CBORSerializer
@@ -765,7 +822,7 @@ else:
 
     class UBJSONObjectSerializer(object):
 
-        NAME = u'ubjson'
+        NAME = 'ubjson'
 
         UBJSON_MODULE = ubjson
 
@@ -833,7 +890,7 @@ else:
 
     class UBJSONSerializer(Serializer):
 
-        SERIALIZER_ID = u"ubjson"
+        SERIALIZER_ID = "ubjson"
         """
         ID used as part of the WebSocket subprotocol name to identify the
         serializer with WAMP-over-WebSocket.
@@ -845,7 +902,7 @@ else:
         handshake identify the serializer with WAMP-over-RawSocket.
         """
 
-        MIME_TYPE = u"application/ubjson"
+        MIME_TYPE = "application/ubjson"
         """
         MIME type announced in HTTP request/response headers when running
         WAMP-over-Longpoll HTTP fallback.
@@ -860,7 +917,7 @@ else:
             """
             Serializer.__init__(self, UBJSONObjectSerializer(batched=batched))
             if batched:
-                self.SERIALIZER_ID = u"ubjson.batched"
+                self.SERIALIZER_ID = "ubjson.batched"
 
     ISerializer.register(UBJSONSerializer)
     SERID_TO_SER[UBJSONSerializer.SERIALIZER_ID] = UBJSONSerializer
@@ -882,7 +939,7 @@ if _HAS_FLATBUFFERS:
 
     class FlatBuffersObjectSerializer(object):
 
-        NAME = u'flatbuffers'
+        NAME = 'flatbuffers'
 
         FLATBUFFERS_MODULE = flatbuffers
 
@@ -935,7 +992,7 @@ if _HAS_FLATBUFFERS:
 
     class FlatBuffersSerializer(Serializer):
 
-        SERIALIZER_ID = u"flatbuffers"
+        SERIALIZER_ID = "flatbuffers"
         """
         ID used as part of the WebSocket subprotocol name to identify the
         serializer with WAMP-over-WebSocket.
@@ -947,7 +1004,7 @@ if _HAS_FLATBUFFERS:
         handshake identify the serializer with WAMP-over-RawSocket.
         """
 
-        MIME_TYPE = u"application/x-flatbuffers"
+        MIME_TYPE = "application/x-flatbuffers"
         """
         MIME type announced in HTTP request/response headers when running
         WAMP-over-Longpoll HTTP fallback.
@@ -961,7 +1018,7 @@ if _HAS_FLATBUFFERS:
             """
             Serializer.__init__(self, FlatBuffersObjectSerializer(batched=batched))
             if batched:
-                self.SERIALIZER_ID = u"flatbuffers.batched"
+                self.SERIALIZER_ID = "flatbuffers.batched"
 
     ISerializer.register(FlatBuffersSerializer)
     SERID_TO_SER[FlatBuffersSerializer.SERIALIZER_ID] = FlatBuffersSerializer
